// -*- C++ -*-
/*!
 * @file OutPort.h
 * @brief OutPort class
 * @date $Date: 2007-12-31 03:08:05 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006-2008
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: OutPort.h 706 2008-04-16 00:36:15Z n-ando $
 *
 */

#ifndef OutPort_h
#define OutPort_h

#include <rtm/BufferBase.h>
#include <rtm/RingBuffer.h>
#include <rtm/OutPortBase.h>
#include <rtm/PortCallBack.h>
#include <rtm/RTC.h>
#include <iostream>

namespace RTC
{
  /*!
   * @if jp
   *
   * @class OutPort
   *
   * @brief OutPort ev[gNX
   * 
   * OutPort ev[gNXAev[g̈ƂāAOutPort
   * f[^^ DataType, OutPort̃obt@^ BufferType ƂB
   *
   * @since 0.2.0
   *
   * @else
   *
   * @brief OutPort template class
   * 
   * This is the OutPort template class. 
   * The data type "DateType" supported by OutPort and the buffer type
   * "BufferType" are used as template arguments.
   *
   * @since 0.2.0
   *
   * @endif
   */
  template <class DataType,
	    template <class DataType> class Buffer = RingBuffer >
  class OutPort
    : public OutPortBase,
      public Buffer<DataType>
  {
  public:
    /*!
     * @if jp
     *
     * @brief RXgN^
     *
     * RXgN^
     *
     * @param name |[g
     * @param value ̃|[gɃoChf[^ϐ
     * @param length obt@(ftHgl:8)
     *
     * @else
     *
     * @brief Constructor
     *
     * Constructor
     *
     * @param name Port's name
     * @param value Data variable bound to this port
     * @param length Buffer length (The default value:8)
     *
     * @endif
     */
    OutPort(const char* name, DataType& value, long int length = 8)
      : OutPortBase(name), Buffer<DataType>(length),
	m_value(value),
	m_timeoutTick(1000), // timeout tick: 1ms
	m_readBlock(false), m_readTimeout(0),
	m_writeBlock(false), m_writeTimeout(0),
	m_OnWrite(NULL), m_OnWriteConvert(NULL),
	m_OnRead(NULL),  m_OnReadConvert(NULL),
	m_OnOverflow(NULL), m_OnUnderflow(NULL)
    {
    }
    
    /*!
     * @if jp
     *
     * @brief fXgN^
     * 
     * fXgN^
     * 
     * @else
     *
     * @brief Destructor
     * 
     * Destructor
     * 
     * @endif
     */
    virtual ~OutPort()
    {
    }
    
    /*!
     * @if jp
     *
     * @brief f[^
     *
     * |[gփf[^ށB
     *
     * - R[obNt@N^ OnWrite ZbgĂꍇA
     *   OutPort ێobt@ɏޑO OnWrite Ă΂B
     * - OutPort ێobt@I[o[t[ołobt@łA
     *   AލۂɃobt@I[o[t[oꍇA
     *   R[obNt@N^ OnOverflow Ă΂B
     * - R[obNt@N^ OnWriteConvert ZbgĂꍇA
     *   obt@ݎɁA OnWriteConvert  operator() ̖߂l
     *   obt@ɏ܂B
     *
     * @param value ݑΏۃf[^
     *
     * @return ݏ(ݐ:trueAݎs:false)
     *
     * @else
     *
     * @brief Write data
     *
     * Write data in the port.
     *
     * - When callback functor OnWrite is already set, OnWrite will be
     *   invoked before writing into the buffer held by OutPort.
     * - When the buffer held by OutPort can detect the overflow,
     *   and when it detected the overflow at writing, callback functor
     *   OnOverflow will be invoked.
     * - When callback functor OnWriteConvert is already set, 
     *   the return value of operator() of OnWriteConvert will be written
     *   into the buffer at the writing.
     *
     * @param value The target data for writing
     *
     * @return Writing result (Successful:true, Failed:false)
     *
     * @endif
     */
    virtual bool write(const DataType& value)
    {
      if (m_OnWrite != NULL)
	{
	  (*m_OnWrite)(value);
	}
      
      long int timeout = m_writeTimeout;
      
      timeval tm_cur, tm_pre;
      ACE_Time_Value tt;
      tt = ACE_OS::gettimeofday();
      tm_pre = tt.operator timeval();
      
      // blocking and timeout wait
      long int count(0);
      while (m_writeBlock && this->isFull())
	{
	  if (m_writeTimeout < 0) 
	    {
	      usleep(m_timeoutTick);
	      continue;
	    }
	  
	  // timeout wait
	  ACE_Time_Value tt;
	  tt = ACE_OS::gettimeofday();
	  tm_cur = tt.operator timeval();
	  long int sec (tm_cur.tv_sec  - tm_pre.tv_sec);
	  long int usec(tm_cur.tv_usec - tm_pre.tv_usec);
	  
	  timeout -= (sec * usec_per_sec + usec);
	  if (timeout < 0) break;
	  
	  tm_pre = tm_cur;
	  usleep(m_timeoutTick);
	  ++count;
	}
      
      if (this->isFull())
	{
	  if (m_OnOverflow != NULL)
	    (*m_OnOverflow)(value);
	  return false;
	}
      
      if (m_OnWriteConvert == NULL) 
	{
	  this->put(value);
	}
      else
	{
	  this->put((*m_OnWriteConvert)(value));
	}
      notify();
      return true;
    }
    
    /*!
     * @if jp
     *
     * @brief f[^
     *
     * |[gփf[^ށB
     * oChꂽϐɐݒ肳ꂽl|[gɏށB
     *
     * @return ݏ(ݐ:trueAݎs:false)
     *
     * @else
     *
     * @brief Write data
     *
     * Write data to the port.
     * Write the value, which was set to the bound variable, to the port.
     *
     * @return Writing result (Successful:true, Failed:false)
     *
     * @endif
     */
    bool write()
    {
      return write(m_value);
    }
    
    /*!
     * @if jp
     *
     * @brief f[^
     *
     * |[gփf[^ށB
     * ݒ肳ꂽl|[gɏށB
     *
     * @param value ݑΏۃf[^
     *
     * @return ݏ(ݐ:trueAݎs:false)
     *
     * @else
     *
     * @brief Write data
     *
     * Write data to the port.
     * Write the set value to the port.
     *
     * @param value The target data for writing
     *
     * @return Writing result (Successful:true, Failed:false)
     *
     * @endif
     */
    bool operator<<(DataType& value)
    {
      return write(value);
    }
    
    /*!
     * @if jp
     *
     * @brief f[^ǂݏo
     *
     * DataPort lǂݏo
     *
     * - R[obNt@N^ OnRead ZbgĂꍇA
     *   DataPort ێobt@ǂݏoO OnRead Ă΂B
     * - DataPort ێobt@A_[t[ołobt@ŁA
     *   AǂݏoۂɃobt@A_[t[oꍇA
     *   R[obNt@N^ OnUnderflow Ă΂B
     * - R[obNt@N^ OnReadConvert ZbgĂꍇA
     *   obt@ݎɁAOnReadConvert  operator() ̖߂l
     *   read()̖߂lƂȂB
     * - setReadTimeout() ɂǂݏõ^CAEgݒ肳ĂꍇA
     *   obt@A_[t[Ԃ܂Ń^CAEgԂ҂A
     *   OnUnderflow ZbgĂ΂ĂяoĖ߂
     *
     * @param value ǂݏof[^
     *
     * @return ǂݏos(ǂݏo:trueAǂݏos:false)
     *
     * @else
     *
     * @brief Readout the data
     *
     * Readout the value from DataPort
     *
     * - When Callback functor OnRead is already set, OnRead will be invoked
     *   before reading from the buffer held by DataPort.
     * - When the buffer held by DataPort can detect the underflow,
     *   and when it detected the underflow at reading, callback functor
     *   OnUnderflow will be invoked.
     * - When callback functor OnReadConvert is already set, the return value of
     *   operator() of OnReadConvert will be the return value of read().
     * - When timeout at the time of reading is already set by setReadTimeout(),
     *   it waits for only timeout period until the state of the buffer underflow
     *   is released, and if OnUnderflow is already set, this will be invoked to
     *   return.
     *
     * @param value Readout data
     *
     * @return Readout result (Successful:true, Failed:false)
     *
     * @endif
     */
    bool read(DataType& value)
    {
      if (m_OnRead != NULL) (*m_OnRead)();      
      
      long int timeout = m_readTimeout;
      timeval tm_cur, tm_pre;
      ACE_Time_Value tt;
      tt = ACE_OS::gettimeofday();
      tm_pre = tt.operator timeval();
      
      // blocking and timeout wait
      while (m_readBlock && this->isEmpty())
	{
	  if (m_readTimeout < 0)
	    {
	      usleep(m_timeoutTick);
	      continue;
	    }
	  
	  // timeout wait
	  ACE_Time_Value tt;
	  tt = ACE_OS::gettimeofday();
	  tm_cur = tt.operator timeval();
	  long int sec (tm_cur.tv_sec  - tm_pre.tv_sec);
	  long int usec(tm_cur.tv_usec - tm_pre.tv_usec);
	  
	  timeout -= (sec * usec_per_sec + usec);
	  if (timeout < 0) break;
	  
	  tm_pre = tm_cur;
	  usleep(m_timeoutTick);
	}
      
      if (this->isEmpty())
	{
	  if (m_OnUnderflow != NULL)
	    {
	      value = (*m_OnUnderflow)();
	      return false;
	    }
	  else
	    {
	      return false;
	    }
	}
      
      if (m_OnReadConvert == NULL) 
	{
	  value = this->get();
	  return true;
	}
      else
	{
	  value = (*m_OnReadConvert)(this->get());
	  return true;
	}
      // never comes here
      return false;
    }
    
    /*!
     * @if jp
     *
     * @brief f[^ǂݏõubN[h̐ݒ
     *
     * ǂݏoɑ΂ăubN[hݒ肷B
     * ubN[hw肵ꍇAǂݏof[^M邩^CAEg
     * ܂ŁAread() \bȟĂтubNB
     *
     * @param block ubN[htO
     *
     * @else
     *
     * @brief Set blocking mode of the data read processing
     *
     * Set the blocking mode for the readout.
     * When the block mode is specified, the invoke of read() method is
     * blocked until the readout data is received or timeout occurs.
     *
     * @param block Flag of blocking mode
     *
     * @endif
     */
    void setReadBlock(bool block)
    {
      m_readBlock = block;
    }
    
    /*!
     * @if jp
     *
     * @brief f[^ݏ̃ubN[h̐ݒ
     *
     * ݏɑ΂ăubN[hݒ肷B
     * ubN[hw肵ꍇAobt@ɏޗ̈悪ł邩
     * ^CAEg܂ write() \bȟĂтubNB
     *
     * @param block ubN[htO
     *
     * @else
     *
     * @brief Set blocking mode of the data writing processing
     *
     * Set the blocking mode for the writing.
     * When the block mode is specified, the invoke of write() method is
     * blocked until the area written into the buffer can be used or timeout
     * occurs.
     *
     * @param block Flag of blocking mode
     *
     * @endif
     */
    void setWriteBlock(bool block)
    {
      m_writeBlock = block;
    }
    
    /*!
     * @if jp
     *
     * @brief ǂݏõ^CAEgԂ̐ݒ
     * 
     * read() ̃^CAEgԂ usec Őݒ肷B
     * read() ̓ubN[hłȂ΂ȂȂB
     *
     * @param timeout ^CAEg [usec]
     *
     * @else
     *
     * @brief Set timeout of the data read processing
     * 
     * Set the timeout period of read() with usec.
     * read() must be a block mode.
     *
     * @param timeout Timeout period[usec]
     *
     * @else
     *
     * @endif
     */
    void setReadTimeout(long int timeout)
    {
      m_readTimeout = timeout;
    }
    
    /*!
     * @if jp
     *
     * @brief ݏ̃^CAEgԂ̐ݒ
     * 
     * write() ̃^CAEgԂ usec Őݒ肷B
     * write() ̓ubN[hłȂ΂ȂȂB
     *
     * @param timeout ^CAEg [usec]
     *
     * @else
     *
     * @brief Set timeout of the data writing processing
     * 
     * Set the timeout period of write() with usec.
     * write() must be a block mode.
     *
     * @param timeout Timeout period[usec]
     *
     * @endif
     */
    void setWriteTimeout(long int timeout)
    {
      m_writeTimeout = timeout;
    }
    
    /*!
     * @if jp
     *
     * @brief OnWrite R[obN̐ݒ
     *
     * f[^ݒOɌĂ΂ OnWrite R[obNt@N^ݒ肷B
     *
     * @param on_write OnWrite R[obNt@N^
     *
     * @else
     *
     * @brief Set OnWrite callback
     *
     * Set OnWrite callback functor invoked immediately before data is written.
     *
     * @param on_write OnWrite callback functor
     *
     * @endif
     */
    inline void setOnWrite(OnWrite<DataType>* on_write)
    {
      m_OnWrite = on_write;
    }
    
    /*!
     * @if jp
     *
     * @brief OnWriteConvert R[obN̐ݒ
     *
     * f[^ݎɌĂ΂ OnWriteConvert R[obNt@N^ݒ
     * B
     * ̃R[obN֐̏ʂ܂B
     * ̂ߏ݃f[^̃tB^O\ƂȂB
     *
     * @param on_wconvert OnWriteConvert R[obNt@N^
     *
     * @else
     *
     * @brief Set OnWriteConvert callback
     *
     * Set OnWriteConvert callback functor invoked when data is written.
     * The processing result of this callback function is written.
     * Therefore, filtering of the writing data will be possible.
     *
     * @param on_wconvert OnWriteConvert callback functor
     *
     * @endif
     */
    inline void setOnWriteConvert(OnWriteConvert<DataType>* on_wconvert)
    {
      m_OnWriteConvert = on_wconvert;
    }
    
    /*!
     * @if jp
     *
     * @brief OnOverflow R[obN̐ݒ
     *
     * obt@tɂf[^݂łȂꍇɌĂяo OnOverflow
     * R[obNt@N^ݒ肷B
     *
     * @param on_overflow OnOverflow R[obNt@N^
     *
     * @else
     *
     * @brief Set OnOverflow callback
     *
     * Set OnOverflow callback functor, which is invoked when data cannot be
     * written because of the buffer full.
     *
     * @param on_overflow OnOverflow callback functor
     *
     * @endif
     */
    inline void setOnOverflow(OnOverflow<DataType>* on_overflow)
    {
      m_OnOverflow = on_overflow;
    }
    
    /*!
     * @if jp
     *
     * @brief OnRead R[obN̐ݒ
     *
     * f[^ǂݏoOɌĂяo OnRead R[obNt@N^ݒ
     * B
     *
     * @param on_read OnRead R[obNt@N^
     *
     * @else
     *
     * @brief Set OnRead callback
     *
     * Set OnWrite callback functor invoked immediately before data is readout.
     *
     * @param on_read OnRead callback functor
     *
     * @endif
     */
    inline void setOnRead(OnRead<DataType>* on_read)
    {
      m_OnRead = on_read;
    }
    
    /*!
     * @if jp
     *
     * @brief OnReadConvert R[obN̐ݒ
     *
     * f[^ǂݏoɌĂ΂ OnReadConvert R[obNt@N^ݒ
     * B
     * ̃R[obN֐̏ʂǂݍ܂B
     * ̂ߓǂݍ݃f[^̃tB^O\ƂȂB
     *
     * @param on_rconvert OnReadConvert R[obNt@N^
     *
     * @else
     *
     * @brief Set OnReadConvert callback
     *
     * Set OnReadConvert callback functor invoked when data is readout.
     * The processing result of this callback function is readout.
     * Therefore, filtering of the read data will be possible.
     *
     * @param on_rconvert OnReadConvert callback functor
     *
     * @endif
     */
    inline void setOnReadConvert(OnReadConvert<DataType>* on_rconvert)
    {
      m_OnReadConvert = on_rconvert;
    }
    
    /*!
     * @if jp
     *
     * @brief OnUnderflow R[obN̐ݒ
     *
     * obt@GveBɂǂݏof[^ȂꍇɌĂяo
     * R[obNt@N^ OnUnderflow ݒ肷B
     *
     * @param on_underflow OnUnderflow R[obNt@N^
     *
     * @else
     *
     * @brief Set OnUnderflow callback
     *
     * Set OnUnderflow callback functor, which is invoked when data cannot be
     * readout because of the buffer empty.
     *
     * @param on_underflow OnUnderflow callback functor
     *
     * @endif
     */
    inline void setOnUnderflow(OnUnderflow<DataType>* on_underflow)
    {
      m_OnUnderflow = on_underflow;
    }    
    
  private:
    /*!
     * @if jp
     * @brief oCh T ^̕ϐւ̎Q
     * @else
     * @brief The reference to type-T variable that is bound.
     * @endif
     */
    DataType& m_value;
    
    /*!
     * @if jp
     * @brief ^CAEg̃|[O [usec]
     * @else
     * @brief Polling cycle of time-out [usec]
     * @endif
     */
    long int m_timeoutTick;
    
    /*!
     * @if jp
     * @brief ǂݍݏ̃ubNEubN[htO
     * @else
     * @brief Flag of read()'s blocking/non-blocking mode
     * @endif
     */
    bool m_readBlock;
    
    /*!
     * @if jp
     * @brief ǂݍݏ̃^CAEg [usec]
     * @else
     * @brief Timeout of read() [usec]
     * @endif
     */
    long int m_readTimeout;
    
    /*!
     * @if jp
     * @brief ݏ̃ubNEubN[htO
     * @else
     * @brief Flag of write()'s blocking/non-blocking mode
     * @endif
     */
    bool m_writeBlock;
    
    /*!
     * @if jp
     * @brief ݏ̃^CAEg [usec]
     * @else
     * @brief Timeout of write() [usec]
     * @endif
     */
    long int m_writeTimeout;
    
    /*!
     * @if jp
     * @brief OnWrite R[obNt@N^ւ̃|C^
     * @else
     * @brief Pointer to OnWrite callback functor
     * @endif
     */
    OnWrite<DataType>* m_OnWrite;
    
    /*!
     * @if jp
     * @brief OnWriteConvert R[obNt@N^ւ̃|C^
     * @else
     * @brief Pointer to OnWriteConvert callback functor
     * @endif
     */
    OnWriteConvert<DataType>* m_OnWriteConvert;
    
    /*!
     * @if jp
     * @brief OnRead R[obNt@N^ւ̃|C^
     * @else
     * @brief Pointer to OnRead callback functor
     * @endif
     */
    OnRead<DataType>* m_OnRead;
    
    /*!
     * @if jp
     * @brief OnReadConvert R[obNt@N^ւ̃|C^
     * @else
     * @brief Pointer to OnReadConvert callback functor
     * @endif
     */
    OnReadConvert<DataType>* m_OnReadConvert;
    
    /*!
     * @if jp
     * @brief OnOverflow R[obNt@N^ւ̃|C^
     * @else
     * @brief Pointer to OnOverflow callback functor
     * @endif
     */
    OnOverflow<DataType>* m_OnOverflow;
    
    /*!
     * @if jp
     * @brief OnUnderflow R[obNt@N^ւ̃|C^
     * @else
     * @brief Pointer to OnUnderflow callback functor
     *
     * @endif
     */
    OnUnderflow<DataType>* m_OnUnderflow;
    
    static const long int usec_per_sec = 1000000;
  };
}; // namespace RTC

#endif // OutPort_h
